/* -*- Mode: JDE; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * This file is part of the debugger and core tools for the AspectJ(tm)
 * programming language; see http://aspectj.org
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * either http://www.mozilla.org/MPL/ or http://aspectj.org/MPL/.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is AspectJ.
 *
 * The Initial Developer of the Original Code is Xerox Corporation. Portions
 * created by Xerox Corporation are Copyright (C) 1999-2002 Xerox Corporation.
 * All Rights Reserved.
 */
package org.aspectj.util;

import java.io.*;
import java.util.*;

// XXX wes remove this class as unused, inverse dependency
public class Ajc {

    public Ajc() {
        check();
        reload();
    }
    
    private String cmdLine = "";
    private PrintStream out = System.out;
    private PrintStream err = System.err;
    private InputStream in = System.in;
    private BufferedReader input = null;
    private final String prompt = "> ";
    private Map arguments = new HashMap();
    private Collection files = new HashSet();
    private void removeFile(String file) {
        files.remove(file);
    }
    private final Integer EMPTY = new Integer(0);
    private void setArgument(String arg, String val) {
        arguments.put(arg, val);
    }
    private void setArgument(String arg) {
        arguments.put(arg, EMPTY);
    }
    private void unsetArgument(String arg) {
        arguments.remove(arg);
    }

    public void realMain(String[] args) {
        parse(args);
        open();
        String cmd;
        while (!quit(cmd = input())) {
            process(cmd);
        }
        close();
        save();
    }

    private void check() {
        try {
            Class.forName("org.aspectj.tools.ajc.Main");
        } catch (ClassNotFoundException e) {
            err.println("Cannot find main class");
            exit();
        }
    }

    private String usage() {
        return "Look at the source!";
    }

    private void process(String cmd) {
        cmd = (cmd != null ? cmd : "").trim();
        if (thread != null && "".equals(cmd)) {
            out(prompt());
        }
        if (cmd.startsWith("ajc") || cmd.startsWith("$")) {
            int iwhite = cmd.indexOf(" ");
            if (iwhite != -1) {
                process(cmd.substring(iwhite));
            }
            String cmdLine = "ajc";
            String[] args = args();
            for (int i = 0; i < args.length; i++) {
                cmdLine += " " + args[i];
            }
            outln(newline + "\t" + cmdLine + newline);
            while(compile(args)) {}
            Thread.yield();
        } else if (cmd.startsWith("help")) {
            outln(usage());
        } else if ("files".equals(cmd)) {
            outln(filesString());
        } else if ("args".equals(cmd)) {
            outln(argumentsString());
        } else if (cmd.startsWith("clear")) {
            boolean clearArgs = false;
            boolean clearFiles = false;
            if (cmd.length() > "clear".length()) {
                String rest = cmd.substring("clear".length()+1).trim();
                System.out.println("rest=" + rest);
                if ("files".equals(rest)) clearFiles = true;
                if ("args".equals(rest)) clearArgs = true;
            } else {
                clearArgs = true;
                clearFiles = true;
            }
            if (clearArgs) arguments = new HashMap();
            if (clearFiles) files = new HashSet();
        } else if ("save".equals(cmd)) {
            save();
        } else if ("load".equals(cmd)) {
            reload();
        } else {
            update(cmd.trim());
        }
            
    }

    private void update(String cmd) {
        files.addAll(removeArguments(cmd));
    }

    private void update(List args) {
        files.addAll(removeArguments(args));
    }

    private final static String quote = "\"";
    private List removeArguments(String cmd) {
        return removeArguments(new ArgumentScanner().scanToList(cmd));
    }
    private List removeArguments(List args) {
        List rest = new Vector();
        Iterator iter = args.iterator();
        String arg;
        String foundArg = null;
        boolean remove = false;
        while (iter.hasNext()) {
            arg = (""+iter.next()).trim();
            if (arg.startsWith("!")) {
                remove = true;
                if (arg.length() == 1 && iter.hasNext()) {
                    arg = iter.next() + "";
                } else {
                    arg = arg.substring(1);
                }
            } else {
                remove = false;
            }
            if (arg.startsWith("-")) {
                foundArg = arg;
                if (!remove) {
                    setArgument(foundArg);
                } else {
                    unsetArgument(foundArg);
                }
            } else {
                if (foundArg == null) {
                    if (remove) {
                        removeFile(arg);
                    } else {
                        rest.add(arg);
                    }
                } else {
                    setArgument(foundArg, arg);
                } 
                foundArg = null;
            }
        }
        return rest;
    }

    

    private boolean quit(String cmd) {
        return "quit".equals(cmd) || "exit".equals(cmd);
    }

    private String input() {
        try {
            return input_();
        } catch (Throwable t) {
            handle(t);
        }
        return "quit";
    }
    private void open() {
        try {
            input = new BufferedReader(new InputStreamReader(in));
        } catch (Throwable t) {
            handle(t, true);
        }
    }
    private void close() {
        try {
            if (input != null) input.close();
        } catch (Throwable t) {
            handle(t, true);
        }
    }
    private String input_() throws IOException {
        out(prompt());
        return input.readLine();
    }

    private String[] args() {
        String[] args = new String[argsList().size() + filesList().size()];
        int i = 0;
        Iterator iter;
        iter = argsList().iterator();
        while (iter.hasNext()) args[i++] = iter.next() + "";
        iter = filesList().iterator();
        while (iter.hasNext()) args[i++] = iter.next() + "";
        return args;
    }

    private List filesList() {
        List list = new Vector();
        Iterator iter = files.iterator();
        while (iter.hasNext()) {
            String next = (""+iter.next()).trim();
            if (!"".equals(next)) list.add(next);
        }
        return list;
    }

    private List argsList() {
        List args = new Vector();
        Iterator iter = arguments.keySet().iterator();
        while (iter.hasNext()) {
            Object arg = iter.next();
            args.add(arg);
            Object val = arguments.get(arg);
            if (val != null && !val.equals(EMPTY)) {
                args.add(val);
            }
        }
        return args;
    }

    private final static int maxArgsLength = 50;
    private final static int maxFilesLength = 50;
    private final static int maxPromptLength = 100;
    private String reduce(String str, int i) {
        if (str.length() > i) return "...";
        return str;
    }
    private String arguments() {
        return reduce(reduce(argumentsString(), maxArgsLength) + ":" + 
                      reduce(filesString(), maxFilesLength),
                      maxPromptLength).trim();
    }

    private String argumentsString() {
        String argsStr = "";
        Iterator iter = arguments.keySet().iterator();
        while (iter.hasNext()) {
            Object arg = iter.next();
            Object val = arguments.get(arg);
            argsStr += arg;
            if (!val.equals(EMPTY)) {
                argsStr += " " + val;
            }
            if (iter.hasNext()) {
                argsStr += ",";
            }
        }
        return argsStr;
    }

    private String filesString() {
        String filesStr = "";
        Iterator iter = files.iterator();
        while (iter.hasNext()) {
            filesStr += iter.next() + (iter.hasNext() ? " " : "");
        }
        return filesStr;
    }

    private String prompt() {
        return arguments() + prompt;
    }

    private boolean time_opt = true;
    private void parse(String[] args) {
        List list = new Vector();
        for (int i = 0; i < args.length; i++) {
            list.add(args[i]);
        }
        update(list);
    }

    private Thread thread;
    private boolean compile(final String[] args) {
        if (thread != null) {
            outln("A compile is already running: " + cmdLine);
            return false;
        }
        try {
            thread = new Thread(new Runnable() {
                    public void run() {
                        long start = -1;
                        try {
                            start = System.currentTimeMillis();
                            Class mainClass = Class.forName("org.aspectj.tools.ajc.Main");
                            mainClass.getMethod("compile", new Class[]{String[].class}).invoke
                                (mainClass.newInstance(),new Object[]{args});
                        } catch (Throwable t) {
                            handle(t);
                        } finally {
                            thread = null;
                            long stop = System.currentTimeMillis();
                            if (time_opt) outln("total time: " + (((double)stop-start)/1000) + "ms.");
                            out(prompt());
                        }
                    }
                });
            thread.start();
            Thread.yield();
        } catch (Throwable t) {
            handle(t);
        }
        return false;
    }

    private final static String newline = System.getProperty("line.separator");

    private void handle(Throwable t, boolean exit) {
        errln("Internal error:");
        t.printStackTrace(err);
        if (exit) exit();
    }

    private void handle(Throwable t) {
        handle(t, false);
    }

    private void exit() {
        save();
        System.exit(-1);
    }

    private void out(Object msg) {
        out.print(msg);
    }

    private void outln(Object msg) {
        out.println(msg);
    }

    private void err(Object msg) {
        err.print(msg);
    }

    private void errln(Object msg) {
        err.println(msg);
    }

    private boolean verbose = true;
    private void log(Object msg) {
        if (verbose) outln("[" + msg + "]");
    }

    public static void main(String[] args) {
        new Ajc().realMain(args);
    }

    private static final String saveFileName = ".ajcoptions";
    private void save() {
        try {
            File saveFile = saveFile();
            if (saveFile == null) return;
            FileOutputStream ostream = new FileOutputStream(saveFile);
            ObjectOutputStream p = new ObjectOutputStream(ostream);
            p.writeObject(arguments);
            p.writeObject(files);
            p.flush();
            ostream.close();
        } catch (Throwable t) {
            handle(t);
        }
    }
    
    private File saveFile() {
        String home = System.getProperty("user.home");
        File homeDir = new File(home);
        File saveFile = null;
        if (homeDir != null && homeDir.exists() && homeDir.isDirectory()) {
            saveFile = new File(homeDir, saveFileName);
        }
        if (saveFile == null) {
            saveFile = new File(saveFileName);
        }
        return saveFile;
    }
    
    private void reload() {
        try {
            File saveFile = saveFile();
            if (saveFile == null || !saveFile.exists()) return;
            FileInputStream istream = new FileInputStream(saveFile);
            ObjectInputStream p = new ObjectInputStream(istream);
            arguments = (Map) p.readObject();
            files = (Collection) p.readObject();
            istream.close();
        } catch (Throwable t) {
            handle(t);
        }
    }
}
